`include "define.h"

module ALU(AIn, BIn, Out, ALUCtrl);

input   [`DATA] AIn, BIn;
output  [`DATA] Out;
input   [`ALUC] ALUCtrl;

reg     [`DATA] Out;
integer         atmp, btmp;

always @(ALUCtrl or AIn or BIn) begin
   if (`dbg) $display(" alu> A %d  B %d   op %b", AIn, BIn, ALUCtrl);
   case (ALUCtrl)
      `ALUAND: begin
                  if (`vbs) $display(" ALUAND: %d & %d", AIn, BIn);
                  Out = AIn & BIn;
               end
      `ALUOR: begin
                  if (`vbs) $display(" ALUOR: %d | %d", AIn, BIn);
                  Out = AIn | BIn;
               end
      `ALUSHL: begin
                  if (`vbs) $display(" ALUSHL: %d << %d", AIn, BIn[10:6]);
                  Out = AIn << BIn[10:6];
               end
      `ALUSHR: begin
                  if (`vbs) $display(" ALUSHR: %d >> %d", AIn, BIn[10:6]);
                  Out = AIn >> BIn[10:6];
               end
      `ALUADD: begin   // add 2-complement numbers
                  if (AIn[`DATAWIDTH-1] == 1)  // negative A
                     atmp = -(~AIn + 1);
                  else
                     atmp = AIn;
                  if (BIn[`DATAWIDTH-1] == 1)  // negative B
                     btmp = -(~BIn + 1);
                  else
                     btmp = BIn;
                  atmp = atmp + btmp;
                  Out = atmp;
                  if (`dbg) $display(" ALUADD: %d + %d = %d (%b)", AIn, BIn, atmp, Out);
               end
      `ALUSUB: begin   // subtract 2-complement numbers
                  if (AIn[`DATAWIDTH-1] == 1)  // negative A
                     atmp = -(~AIn + 1);
                  else
                     atmp = AIn;
                  if (BIn[`DATAWIDTH-1] == 1)  // negative B
                     btmp = -(~BIn + 1);
                  else
                     btmp = BIn;
                  atmp = atmp - btmp;
                  Out = atmp;
                  if (`vbs) $display(" ALUSUB: %d - %d = %d (%b)", AIn, BIn, atmp, Out);
               end
      `ALUMUL: begin   // multiply 2-complement numbers
                  if (AIn[`DATAWIDTH-1] == 1)  // negative A
                     atmp = -(~AIn + 1);
                  else
                     atmp = AIn;
                  if (BIn[`DATAWIDTH-1] == 1)  // negative B
                     btmp = -(~BIn + 1);
                  else
                     btmp = BIn;
                  atmp = atmp * btmp;
                  Out = atmp;
                  if (`vbs) $display(" ALUMUL: %d * %d = %d (%b)", AIn, BIn, atmp, Out);
               end
      default: begin
                  $display(" ALU> ERROR: Illeagal ALUCtrl %b", ALUCtrl);
                  $finish;
               end
   endcase
end

endmodule /* ALU */

module ALUControl(ALUop, OPcode, Func, ALUCtrl);

input   [`ALUO] ALUop;
input   [5:0]   OPcode;
input   [5:0]   Func;
output  [`ALUC] ALUCtrl;

reg     [`ALUC] ALUCtrl;

initial
   ALUCtrl = `ALUCTRLWIDTH'b0;

always @(ALUop or OPcode or Func) begin
   if (ALUop == `ITYPE || ALUop == `BRANCH) begin
      casex ({ALUop,OPcode})
         {`ITYPE,`ADDI}:  ALUCtrl = `ALUADD;   // R-Type: add
         {`BRANCH,6'bx}:  ALUCtrl = `ALUADD;   // J-Type: add
         default: begin
                     $display(" ALUControl> ERROR: Illeagal Input {%b,%b}",
                                   ALUop, OPcode);
                     $finish;
                  end
      endcase
   end
   else begin
      casex ({ALUop,Func})
         {`MEMACC,6'bx}: ALUCtrl = `ALUADD;   // Load / store
         {`RTYPE,`AND}:  ALUCtrl = `ALUAND;   // R-Type: and
         {`RTYPE,`OR}:   ALUCtrl = `ALUOR;    // R-Type: or
         {`RTYPE,`SLL}:  ALUCtrl = `ALUSHL;   // R-Type: shl
         {`RTYPE,`SRL}:  ALUCtrl = `ALUSHR;   // R-Type: shr
         {`RTYPE,`ADD}:  ALUCtrl = `ALUADD;   // R-Type: add
         {`RTYPE,`SUB}:  ALUCtrl = `ALUSUB;   // R-Type: sub
         {`RTYPE,`MUL}:  ALUCtrl = `ALUMUL;   // R-Type: mul
         default: begin
                     $display(" ALUControl> ERROR: Illeagal Input {%b,%b}",
                                   ALUop, Func);
                     $finish;
                  end
      endcase
   end
end

endmodule /* ALUControl */

