`include "define.h"

/*-------------------------------------------------------------------------
 * hazard detector : When data hazard is detected, stall the pipeline 
                     Also, determine forwarding control for hazards
 *------------------------------------------------------------------------*/
module HazardDetection( RegDstIDEX, RegRWIDEX, RegDstEXMEM, RegRWEXMEM,
			RegDstMEMWB, RegRWMEMWB, OPcodeIFID, SrcRegIFID, 
			DstRegIDEX, DstRegEXMEM, DstRegMEMWB, Stall, Reset, Clk, 
			ALUinA, ALUinB, BranchControl, OPcodeIDEX, OPcodeEXMEM);

input			RegDstIDEX, RegRWIDEX;
input			RegDstEXMEM, RegRWEXMEM;
input			RegDstMEMWB, RegRWMEMWB;
input [`OPCODEFIELD]	OPcodeIFID;
input [`OPCODEFIELD]	OPcodeIDEX;
input [`OPCODEFIELD]	OPcodeEXMEM;
input [`SRCREGFIELD]	SrcRegIFID;
input [`DSTREGFIELD]	DstRegIDEX, DstRegEXMEM, DstRegMEMWB;

output			Stall;

output [2:0]		ALUinA;
output [2:0]		ALUinB;
output 			BranchControl;

input 			Reset;
input			Clk;

reg			Stall;
reg [2:0]		ALUinA;
reg [2:0]		ALUinB;
reg 			BranchControl;

integer StallAmt;

initial begin
   Stall = 0;
end

always @( negedge Clk ) begin 
   if ( Reset == 1 ) begin
      Stall = 0;
   end
   else begin

      ALUinA = 0;
      //ALUinB = inALUB ? 1 : 0; 
      ALUinB = 0;
      BranchControl = 0;
      StallAmt = 0;

      if ( `dbg ) 
        $display("Hazard Detector> RegRWIDEX=%b DstRegIDEX[15:11]=%b DstRegIDEX[20:16]=%b SrcRegIFID[25:21]=%b SrcRegIFID[20:16]=%b",
                  RegRWIDEX, DstRegIDEX[`DSTRRDFIELD], DstRegIDEX[`DSTIRDFIELD], SrcRegIFID[`SRCRRS1FIELD],
                  SrcRegIFID[`SRCRRS2FIELD] );
            
      /* detect the data hazard between EX/MEM and IF/ID */            
      if ( RegRWEXMEM == 1'b1 ) begin
         if ( RegDstEXMEM == 1'b1 ) begin
            if ( DstRegEXMEM[`DSTRRDFIELD] == SrcRegIFID[`SRCRRS1FIELD] ) begin

            // Port A Conflict on current Instruction
	    // RType 2X previous Instruction, ALU operations 

               if ( (OPcodeIFID == `BEQZ) || (OPcodeIFID == `BNEZ) ) begin
                  BranchControl = 1;		//Branch forward ALUoutEXMEM
                  if (`dbg) begin
                    $display($time, 
                      " >> Hazard detected between EX/MEM(ALU) and IF/ID (Branch Control)");
                    $display("Forwarding path is ALUoutEXMEM to Ztestin");
                  end
               end
               else begin
                  ALUinA = 3;  //Data forward ALUoutMEMWB
                  if (`dbg) begin
                    $display($time, 
                      " >> Hazard between EX/MEM(ALU) and IF/ID (ALU RS1)"); 
                    $display("Forwarding path is LLUoutMEMWB to aluina");
                  end
               end
	    end 
            if ( DstRegEXMEM[`DSTRRDFIELD] == SrcRegIFID[`SRCRRS2FIELD]  &&
                   ( OPcodeIFID == `FSEL || OPcodeIFID == `SW ) )begin

            // Port B Conflict on current Instruction
	    // RType 2X previous Instruction, ALU operations 

               ALUinB = 3;			//Data forward ALUoutMEMWB
               if (`dbg) begin
                 $display($time, 
                   " >> Hazard between EX/MEM(ALU) and IF/ID (ALU Port B)");
                 $display("Forwarding path is ALUoutMEMWB to aluinb");
               end
            end
         end
         else begin

	    // Port A Conflict on current Instruction
            // IType 2X previous Instruction, ADDI or LW

            if ( DstRegEXMEM[`DSTIRDFIELD] == SrcRegIFID[`SRCRRS1FIELD] ) begin

               if (OPcodeEXMEM == `LW) begin
               	  if ( (OPcodeIFID == `BEQZ) || (OPcodeIFID == `BNEZ) ) begin
                     StallAmt = 1;
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between EX/MEM (LW) and IF/ID (Branch outcome).");
                       $display("Stall needed");
                       $display("No forwarding path needed after stall");
                     end
                  end
                  else begin	
                     ALUinA = 1;
                     if (`dbg) begin//Data forward MD
                       $display($time, 
                         " >> Hazard between EX/MEM (LW) and IF/ID (ALU RS1)"); 
                       $display("Forwarding path is MDMEMWB to aluina");
                     end
                  end
               end
               if (OPcodeEXMEM == `ADDI) begin
                  if ( (OPcodeIFID == `BEQZ) || (OPcodeIFID == `BNEZ) ) begin
                     BranchControl = 1;		//Branch forward ALUoutEXMEM
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between EX/MEM(ADDI) and IF/ID (Branch outcome).");
                       $display("Forwarding path is ALUoutEXMEM to Ztestin");
                     end
                  end
                  else begin
                     ALUinA = 3;                //Data forward ALUoutMEMWB
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between EX/MEM (ADDI) and IF/ID (ALU RS1)");
                       $display("forwarding path is ALUoutEXMEM to aluina");
                     end
                  end
               end
            end
            if ( DstRegEXMEM[`DSTIRDFIELD] == SrcRegIFID[`SRCRRS2FIELD] &&
                   ( OPcodeIFID == `FSEL || OPcodeIFID == `SW ) ) begin
	
	    // Port B Conflict on current Instruction
            // IType 2X previous Instruction, ADDI or LW

               if (OPcodeEXMEM == `LW) begin
                  ALUinB = 1;			//Data forward MD
                  if (`dbg) begin
                    $display($time, 
                       " >> Hazard between EX/MEM (LW) and IF/ID (ALU RS2 or SW)");
                    $display("Forwarding path is MDMEMWB to aluinb");
                  end
               end
               if (OPcodeEXMEM == `ADDI) begin
                  ALUinB = 3;                   //Data forward ALUoutMEMWB
                  if (`dbg) begin
                    $display($time, 
                      " >> Hazard between EX/MEM (ADDI) and IF/ID (ALU RS2)");
                    $display("Forwarding path is ALUoutMEMWB to aluinb");
                  end
               end
            end
         end
      end

      /* detect the data hazard between ID/EX and IF/ID */
      if ( RegRWIDEX == 1'b1 ) begin
         if ( RegDstIDEX == 1'b1 ) begin

            if ( DstRegIDEX[`DSTRRDFIELD] == SrcRegIFID[`SRCRRS1FIELD] ) begin
            // Port A Conflict on current instruction
            // RType previous Instruction, ALU operations
               if ( (OPcodeIFID == `BEQZ) || (OPcodeIFID == `BNEZ) ) begin
                  StallAmt= 1;
                  BranchControl = 1;
                  if (`dbg) begin
                    $display($time, 
                      " >> Hazard between ID/EX (ALU) and IF/ID (branch outcome)");
                    $display("Stall needed");
                    $display("Forwarding path after stall is ALUoutEXMEM to Ztestin");
                  end
               end
               else begin         
                  ALUinA = 2;          // Data forward ALUoutEXMEM
                  if (`dbg) begin
                    $display($time, 
                      " >> Hazard between ID/EX (ALU) and IF/ID (RS1)");
                    $display("Forwarding path is ALUoutEXMEM to aluina");
                  end
               end
            end
            if ( DstRegIDEX[`DSTRRDFIELD] == SrcRegIFID[`SRCRRS2FIELD]  &&
                    ( OPcodeIFID == `FSEL || OPcodeIFID == `SW ) ) begin
            // Port B Conflict on current Instruction
            // RType previous Instruction, ALU operations
               ALUinB = 2;          // Data forward ALUoutExMem
               if (`dbg) begin
                 $display($time, 
                   " >> Hazard between ID/EX (ALU) and IF/ID (RS2)");
                 $display("Forwarding path is ALUoutEXMEM to aluinb");
               end    
            end
         end
         else begin
            if ( DstRegIDEX[`DSTIRDFIELD] == SrcRegIFID[`SRCRRS1FIELD] ) begin
            // Port A Conflict on current instruction
            // IType previous Instruction, ADDI or LW

               if (OPcodeIDEX == `LW) begin
                  if ( (OPcodeIFID == `BEQZ) || (OPcodeIFID == `BNEZ) ) begin
                     StallAmt = 2;
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between ID/EX (LW) and IF/ID (branch outcome)");
                       $display("Two stalls needed");
                       $display("No forwarding path needed after stalls");
                     end
                  end
                  else begin
                     StallAmt = 1;
                     ALUinA = 1;        // Data forward MD
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between ID/EX (LW) and IF/ID (RS1)");
                       $display("Stall needed");
                       $display("Forwarding path after stall is MDMEMWB to aluina");
                     end
                  end
               end
               else if (OPcodeIDEX == `ADDI) begin
                  if ( (OPcodeIFID == `BEQZ) || (OPcodeIFID == `BNEZ) ) begin
                     StallAmt = 1;
                     BranchControl = 1; // Branch forward ALUoutEXMEM
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between ID/EX (ADDI) and IF/ID (branch outcome)");
                       $display("Stall needed");
                       $display("Forwarding path after stall is ALUoutEXMEM to Ztestin");
                     end
                  end
                  else begin
                     ALUinA = 2;        // Data forward ALUoutEXMEM
                     if (`dbg) begin
                       $display($time, 
                         " >> Hazard between ID/EX (ADDI) and IF/ID (RS1)");
                       $display("Forwarding path is ALUoutEXMEM to aluina");
                     end
                  end
               end
            end
            if ( DstRegIDEX[`DSTIRDFIELD] == SrcRegIFID[`SRCRRS2FIELD] &&
                   ( OPcodeIFID == `FSEL || OPcodeIFID == `SW ) ) begin
            // Port B Conflict on current Instruction
            // IType previous Instruction, ADDI or LW
               if (OPcodeIDEX == `LW) begin
                  if (`dbg) begin
                    $display($time, 
                      " >> Hazard between ID/EX (LW) and IF/ID (RS2)");
                    end
                  if (StallAmt == 0) begin   // Don't write over a StallAmt of 2
                     StallAmt = 1;
                     if (`dbg) begin
                       $display("Stall needed");
                     end
                  end
                  ALUinB = 1;           // Data forward MD
                  if (`dbg) begin
                    $display("Forwarding path is MWMEMWB to aluinb");
                  end
               end
               else if (OPcodeIDEX == `ADDI) begin
                  ALUinB = 2;            // Data forward ALUoutEXMEM
                  if (`dbg) begin
                    $display($time, " >> Hazard between ID/EX (ADDI) and IF/ID (RS2).");
                    $display("Forwarding path is ALUoutEXMEM to aluinb");
                  end
               end
            end
         end
      end

      
      /* detect the data hazard between MEM/WB and IF/ID */ 
      /* There are no distance three hazards with "Write before Read" Register behavior */
      /*
      if ( RegRWMEMWB == 1'b1 ) begin
         if ( RegDstMEMWB == 1'b1 ) begin
            if ( DstRegMEMWB[`DSTRRDFIELD] == SrcRegIFID[`SRCRRS1FIELD] || 
                 ( DstRegMEMWB[`DSTRRDFIELD] == SrcRegIFID[`SRCRRS2FIELD]  &&
                   ( OPcodeIFID == `FSEL || OPcodeIFID == `SW ) ) ) begin
               $display($time, " >> Data hazard detected between MEM/WB and IF/ID.");                             
            end
         end
         else begin
            if ( DstRegMEMWB[`DSTIRDFIELD] == SrcRegIFID[`SRCRRS1FIELD] ||
                 ( DstRegMEMWB[`DSTIRDFIELD] == SrcRegIFID[`SRCRRS2FIELD] &&
                   ( OPcodeIFID == `FSEL || OPcodeIFID == `SW ) ) ) begin
               $display($time, " >> Data hazard detected between MEM/WB and IF/ID.");               
            end
         end
      
      end
      */
      if (StallAmt > 0) begin
        if (`dbg) begin
          $display($time, " >> Data hazard detected. Insert %d bubbles.", StallAmt);
        end
        Stall = 1;
        repeat ( StallAmt ) #`CLOCKPERIOD;
        Stall = 0;
      end
   end
end

endmodule
