//
// misc.v: Mem, Register, OneBitReg, Mux, Add4, SignExt, ZTest, Clock
//
`include "define.h"

module Mem (Addr, DataIn, DataOut, MemRW, Clk);
input   [`DATA] Addr;
input   [`DATA] DataIn;
output  [`DATA] DataOut;
input           MemRW;
input           Clk;

reg     [`BYTE] MEM[0:`MEMSIZE-1];

wire    [`DATA] DataOut = {MEM[Addr],MEM[Addr+1],MEM[Addr+2],MEM[Addr+3]};

always @(negedge Clk) begin
   if (MemRW)
      {MEM[Addr],MEM[Addr+1],MEM[Addr+2],MEM[Addr+3]} = DataIn;
end

endmodule /* Mem */

module Register (RegOut, RegIn, RegRW, Clk);
output  [`DATA] RegOut;
input   [`DATA] RegIn;
input           RegRW;
input           Clk;

parameter delay = 2;

reg     [`DATA] RegBuf;
wire    [`DATA] RegOut = RegBuf;

always @(negedge Clk) begin
   if (RegRW) #delay RegBuf = RegIn;
end

endmodule /* Register */

/*
module OneBitReg (RegOut, RegIn, RegRW, Clk);
output  RegOut;
input   RegIn;
input   RegRW;
input   Clk;

parameter delay = 2;

reg     RegBuf;
wire    RegOut = RegBuf;

always @(negedge Clk) begin
   if (RegRW) #delay RegBuf = RegIn;
end

endmodule */ /* OneBitReg */

module Mux (S, A, B, Z);

input           S;      // Select line
input   [`DATA] A;      // input A
input   [`DATA] B;      // input B
output  [`DATA] Z;      // output

wire    [`DATA] Z = S ? B : A;

endmodule /* Mux */

module FourInMux (S, A, B, C, D, Z);

input   [2:0]   S;      // Select line
input   [`DATA] A;      // input A
input   [`DATA] B;      // input B
input   [`DATA] C;      // input C
input   [`DATA] D;      // input D
output  [`DATA] Z;      // output

wire    [`DATA] Z = S === 0 ? A :
		    S === 1 ? B :
                    S === 2 ? C : D;

endmodule /* FourInMux */

/*
module FiveInMux (S, A, B, C, D, E, Z);

input   [3:0]   S;      // Select line
input   [`DATA] A;      // input A
input   [`DATA] B;      // input B
input   [`DATA] C;      // input C 
input   [`DATA] D;      // input D
input   [`DATA] E;      // input E
output  [`DATA] Z;      // output

wire    [`DATA] Z = S === 0 ? A :
		    S === 1 ? B :
		    S === 2 ? C :
		    S === 3 ? D : E;
// 0 -> A
// 1 -> B

endmodule 
*/

module RddrMux (S, A, B, Z);

input           S;      // Select line
input   [`RDDR] A;      // input A
input   [`RDDR] B;      // input B
output  [`RDDR] Z;      // output

wire    [`RDDR] Z = S ? B : A;

endmodule /* RddrMux */

/*
module Mux3Bits( S, A, B, Z );
input           S;      // Select line
input   [2:0] 	A;      // input A
input   [2:0] 	B;      // input B
output  [2:0] 	Z;      // output

wire    [2:0] Z = S ? B : A;

endmodule
*/

/*
module Mux5Bits( S, A, B, Z );
input           S;      // Select line
input   [4:0] 	A;      // input A
input   [4:0] 	B;      // input B
output  [4:0] 	Z;      // output

wire    [4:0] Z = S ? B : A;

endmodule
*/

module Adder (A, B, O);

input   [`DATA] A;      // input
input   [`DATA] B;      // input
output  [`DATA] O;      // output

wire    [`DATA] O = A + B;

endmodule /* Adder */

module Add4 (I, O);

input   [`DATA] I;      // input
output  [`DATA] O;      // output

wire    [`DATA] O = I + 4;

endmodule /* Add4 */

module SignExt (S, I, O);

input           S;
input   [25:0]  I;      // input
output  [`DATA] O;      // output

wire    [`DATA] O = S ? {{6 {I[25]}}, I} : {{16 {I[15]}}, I[15:0]};

endmodule /* SignExt */

module ZTest (S, I, Out);

input           S;      // S ? nez : eqz
input   [`DATA] I;      // input
output          Out;    // output

wire Out = S ? ((I !== 0) ? 1 : 0) : ((I === 0) ? 1 : 0);

endmodule /* ZTest */

module Clock (clk);
output clk;
reg    clk;

initial
   clk = 0;

always
   #(`CLOCKPERIOD / 2)     clk = ~clk;

endmodule /* Clock */

