`include "define.h"

module sdlxDelayedBranch;    // top module

// the pipleline registers
reg	[`DATA] NPCIFID, NPCIDEX, NPCIFIDtemp, NPCIDEXtemp;
reg	[`DATA] IRIFID, IRIDEX, IREXMEM, IRMEMWB, IRIFIDtemp, IRIDEXtemp, IREXMEMtemp, IRMEMWBtemp;
reg	[2:0]	WBIDEX, WBEXMEM, WBMEMWB, WBIDEXtemp, WBEXMEMtemp, WBMEMWBtemp;
reg	[2:0]   MEMIDEX, MEMEXMEM, MEMIDEXtemp, MEMEXMEMtemp;
reg	[4:0]   EXIDEX, EXIDEXtemp;
reg	[`DATA]	A, BIDEX, BEXMEM, ALUoutEXMEM, ALUoutMEMWB, IMMREG, MD;
reg	[`DATA] Atemp, BIDEXtemp, BEXMEMtemp, ALUoutEXMEMtemp, ALUoutMEMWBtemp, IMMREGtemp, MDtemp;
reg		BC, BCtemp;
reg		HALTIDEX, HALTEXMEM, HALTMEMWB, HALTIDEXtemp, HALTEXMEMtemp, HALTMEMWBtemp;

reg	[`DATA] PcReg;
reg 		Reset;
reg	[2:0]	IDEXALUA, IDEXALUAtemp;
reg	[2:0]	IDEXALUB, IDEXALUBtemp;
wire            clk;

wire	[`DATA] seqnpcin, jmpnpcin, pcin, PCBrOut;
wire	[2:0]   wbIFID;
wire	[2:0]   memIFID;
wire 	[4:0]   exIFID;
wire	[2:0]   ALUinA;
wire	[2:0]   ALUinB;
wire		BranchControl;

wire 	[`DATA] rfa, rfb, aluina, aluinb, BForward, ZtestIn, aluout, dmemout,instrmemout,wbrf;
wire		bcin;
wire	[`DATA] iin;
wire		SESel, RegRW;
wire	[`ALUC]	ALUCtrl;
wire    [`RDDR] rd;
wire		Stall;
wire		halt;

integer		stallcount, cyclecount;
integer         addr, size;

// connection of modules
Clock		Clock	( clk );

Control		Control (
	.IR		(IRIFID),
	.SESel		(SESel),
	.WBcontrol	(wbIFID),
	.MEMcontrol	(memIFID),
	.EXcontrol	(exIFID), 
	.Halt		(halt) );

/* Hazard Detection Module outputs control signals for Data Forwarding Mux's and Branch Forwarding Mux
   Hazard Detection Module needs knowledge of IFID, IDEX, and EXMEM Opcodes for Data forwarding */	
HazardDetection HazardDetection (
	.RegDstIDEX	(WBIDEX[1]), 
	.RegRWIDEX	(WBIDEX[0]), 
	.RegDstEXMEM	(WBEXMEM[1]),
	.RegRWEXMEM	(WBEXMEM[0]),
	.RegDstMEMWB	(WBMEMWB[1]),
	.RegRWMEMWB	(WBMEMWB[0]),
	.OPcodeIFID	(IRIFID[`OPCODEFIELD]),
	.OPcodeIDEX	(IRIDEX[`OPCODEFIELD]),
	.OPcodeEXMEM	(IREXMEM[`OPCODEFIELD]),
	.SrcRegIFID	(IRIFID[`SRCREGFIELD]), 
	.DstRegIDEX	(IRIDEX[`DSTREGFIELD]),
	.DstRegEXMEM	(IREXMEM[`DSTREGFIELD]),
	.DstRegMEMWB	(IRMEMWB[`DSTREGFIELD]),
        .Stall 		(Stall),
        .Reset		(Reset),
        .Clk		(clk), 
	.ALUinA		(ALUinA),
	.ALUinB		(ALUinB),
	.BranchControl  (BranchControl) );
	
ALUControl      ALUControl	(`ALUOPCODE, `OPCODEIDEX, `FUNCIDEX, ALUCtrl);
ALU             ALU             (aluina, aluinb, aluout, ALUCtrl);

SignExt         SignExt         (SESel, `OFFSET, iin);
Add4            Add4            (pcin, seqnpcin);
Adder		Adder		(NPCIFID, iin, PCBrOut);

ZTest           ZTest           (`ZSEL, ZtestIn, bcin);

Mux             PCMux           ((BC & `BRAE) | `JMPE, PcReg, NPCIDEX, pcin);
RddrMux         RDMux           (`REGDST, `IRD, `RRD, rd);

/* Data Forwarding Mux's */
FourInMux       ALUAInMux       (IDEXALUA, A, MD, ALUoutEXMEM, ALUoutMEMWB, aluina);
FourInMux       ALUBInMux       (IDEXALUB, BIDEX, MD, ALUoutEXMEM, ALUoutMEMWB, BForward);

/* Branch Forwarding Mux */
Mux             BranchMux       (BranchControl, rfa, ALUoutEXMEM, ZtestIn);

/* ALU in B Mux, controls B register (with data forwarding) versus Immediate value selection */
Mux             ALUBImmInMux    (`ALUINBImm, BForward, IMMREG, aluinb);

Mux             WBMux           (`WBSEL, MD, ALUoutMEMWB, wbrf);

RegFile         RegFile         (`RRS1, `RRS2, rd, wbrf, rfa, rfb, `REGRW, clk);

Mem         	DataMem        (ALUoutEXMEM, BEXMEM, dmemout, `MEMRW, clk);
Mem             InstrMem       (pcin,`DATAWIDTH'b0,instrmemout,1'b0, clk); 


always wait(HALTMEMWB === 1 ) begin
   $display;
   $display("Number of stall cycles due to data hazards is %d", stallcount);
   $display("Total number of clock cycles, including stalls, is: %d", cyclecount);
   $display;
   $display("The result of matrix multiplication : ");
   for ( size=0; size<36; size=size+1 )
      begin
         addr = 288 + (size * 4);
         $display("%d", {DataMem.MEM[addr], DataMem.MEM[addr+1], 
                             DataMem.MEM[addr+2], DataMem.MEM[addr+3]});
      end
   $display;
   $stop;
end

// The following always block implements the pipeline control. The block is executed at each 
// positive clock edge.  The block combines the behavior necessary to complete clock cycle i,
// followed by behavior to begin clock cycle i+1. To complete clock cycle i, stage outputs are
// latched into temporary registers.  To initate cycle i+1 the temporary registers are then copied
// to the actual staging logic registers.  This simulates "master-slave" behavior of the staging
// logic.  
//
// The block is also responsible for implementing pipeline stalls based upon a Stall signal generated
// by the Hazard Detection module.  the Stall signal is asserted on the next negative clock edge
// following the positive edge at which the trailing RAW hazard instruction is lathed into the
// IR (IRIFID).  Stall remains asserted until the next negative clock edge. 
//
// The pipeline assumes Delayed Branching semantics--i.e. the instruction immediately following a
// branch instruction is fetched and executed before the branch takes effect. It is the ressponsibility
// of the compiler to place either a NOP or an instruction not dependent on, or affecting, the branch
// outcome in the branch shadow.
 
always @(posedge clk ) begin
   if ( Reset == 1 ) begin
      PcReg = 4;
      NPCIFID = 4;
      NPCIDEX = 0;
      IRIFID = {InstrMem.MEM[0], InstrMem.MEM[1], InstrMem.MEM[2], InstrMem.MEM[3]};
      IRIDEX = `NOP;
      IREXMEM = `NOP;
      IRMEMWB = `NOP;
      WBIDEX = 0;
      WBEXMEM = 0;
      WBMEMWB = 0;
      MEMIDEX = 0;
      MEMEXMEM = 0;
      EXIDEX = 0;
      A = 0;
      BIDEX = 0;
      BEXMEM = 0;
      ALUoutEXMEM = 0;
      ALUoutMEMWB = 0;
      IMMREG = 0;
      MD = 0;
      BC = 0;
      HALTMEMWB = 0;
      HALTEXMEM = 0;
      HALTIDEX = 0;
   end
   else begin
                           
//Actions to complete the previous cycle:
 

  //Fetch
         if(Stall==0) begin 
           IRIFIDtemp = instrmemout; //next instruction placed in IRIFIDtemp    
           NPCIFIDtemp = seqnpcin;
           PcReg = seqnpcin;
         end
         
  //Decode
         if (Stall==0) begin
             IRIDEXtemp = IRIFID;
	     WBIDEXtemp = wbIFID;     
	     EXIDEXtemp = exIFID;
	     MEMIDEXtemp = memIFID;
	     HALTIDEXtemp = halt;
	     NPCIDEXtemp = PCBrOut;    // NPCIDEX gets computed Branch Target Address
	    /* delay control signals for data forwarding one Stage time */
	     IDEXALUAtemp = ALUinA;	// IDEXALUA is control signal for Data Forward A MUX (four Input)
	     IDEXALUBtemp = ALUinB;	// IDEXALUB is control signal for Data Forward B MUX (four Input) 
	     Atemp = rfa;
	     BIDEXtemp = rfb;               
	     IMMREGtemp = iin;
	     BCtemp = bcin;			// bcin is computed in IDEX stage
         end
         else begin   // stall--insert bubble
            stallcount = stallcount + 1;
	    EXIDEXtemp = 0;
	    HALTIDEXtemp=0;
	    MEMIDEXtemp = 0;
	    WBIDEXtemp=0;
	    IRIDEXtemp=0;
         end
         
  //Execute

      IREXMEMtemp = IRIDEX;
      WBEXMEMtemp = WBIDEX;
      MEMEXMEMtemp = MEMIDEX;
      HALTEXMEMtemp = HALTIDEX;
      BEXMEMtemp = BForward;    // BEXMEM gets data forwarded from Data Forward B MUX (four Input)
      ALUoutEXMEMtemp = aluout;

  //Mem

      IRMEMWBtemp = IREXMEM;
      WBMEMWBtemp = WBEXMEM;
      HALTMEMWBtemp = HALTEXMEM;
      MDtemp = dmemout;
      ALUoutMEMWBtemp = ALUoutEXMEM;
      
//Actions to start the next cycle      
            
//pull forward MEM/WB stage
      IRMEMWB = IRMEMWBtemp;
      WBMEMWB = WBMEMWBtemp;
      HALTMEMWB = HALTMEMWBtemp;
      MD = MDtemp;
      ALUoutMEMWB = ALUoutMEMWBtemp;
      
//pull forward EX/MEM stage
      IREXMEM = IREXMEMtemp;
      WBEXMEM = WBEXMEMtemp;
      MEMEXMEM = MEMEXMEMtemp;
      HALTEXMEM = HALTEXMEMtemp;
      BEXMEM = BEXMEMtemp;    
      ALUoutEXMEM = ALUoutEXMEMtemp;

      
         IRIDEX = IRIDEXtemp;
         WBIDEX = WBIDEXtemp;     
         EXIDEX = EXIDEXtemp;
         HALTIDEX = HALTIDEXtemp;
         NPCIDEX = NPCIDEXtemp;    // NPCIDEX gets computed Branch Target Address
         IDEXALUA = IDEXALUAtemp;  // IDEXALUA is control signal for Data Forward A MUX (four Input)
         IDEXALUB = IDEXALUBtemp;  // IDEXALUB is control signal for Data Forward B MUX (four Input) 
         A = Atemp;
         BIDEX = BIDEXtemp;               
         IMMREG = IMMREGtemp;
         BC = BCtemp;			// BC holds the branch outcome 
         MEMIDEX = MEMIDEXtemp;
            
// pull forward IF/ID if no stall      
      if (Stall==0) begin
         IRIFID = IRIFIDtemp;      
         NPCIFID = NPCIFIDtemp;
      end
   end
   cyclecount = cyclecount +1;
end

// Reset machine
task apply_reset;
begin
   Reset = 1;
   #`CLOCKPERIOD ;
   Reset = 0;
end
endtask

initial begin : load_prog
   integer i;

   
   $readmemb("matmultDelayedBranch.s", InstrMem.MEM);   //Read file containing program into IMEM
   //$readmemb("test.s", InstrMem.MEM);   //Read file containing program into IMEM

   $readmemb("matrix.dat", DataMem.MEM ); // read data into DMEM

   if (`dbg) begin
      $display(" Loaded program:");
      for(i=0; i<400; i=i+4)
               $display("%d", {InstrMem.MEM[i], InstrMem.MEM[i+1], 
                                   InstrMem.MEM[i+2], InstrMem.MEM[i+3]});
   end
   if (`dbg) begin
      $monitor($time, ":%b> pc=%h IRIFID=%h EXIDEX=%b MEMIDEX=%b WBIDEX=%b",
                   clk, PcReg, IRIFID, EXIDEX, MEMIDEX, WBIDEX );
      $monitor( $time, ":%b> pc=%h IRIFID=%h IRIDEX=%h R3=%h R4=%h R23=%h R24=%h", 
                clk, PcReg, IRIFID, IRIDEX,
                RegFile.R3.RegBuf, RegFile.R4.RegBuf, RegFile.R23.RegBuf,
                RegFile.R24.RegBuf );
   end
   
   stallcount=0;
   cyclecount = 0;

   apply_reset;
end

endmodule /* sdlxDelayedBranch */

